@@ -39,32 +39,39 @@ import Data.Ix (Ix)
 #if ! MIN_VERSION_base(4,11,0)
 import Data.Semigroup ((<>))
 #endif
+import Data.WideWord.Word64
 
 import Foreign.Ptr (Ptr, castPtr)
 import Foreign.Storable (Storable (..))
 
-import GHC.Base (Int (..), and#, int2Word#, minusWord#, not#, or#, plusWord#, plusWord2#
-                , quotRemWord2#, subWordC#, timesWord#, timesWord2#, xor#)
+import GHC.Base (Int (..))
 import GHC.Enum (predError, succError)
 import GHC.Exts ((*#), (+#), Int#, State#, ByteArray#, MutableByteArray#, Addr#)
-import GHC.Generics
+import GHC.Generics (Generic)
 import GHC.Real ((%), divZeroError)
-import GHC.Word (Word64 (..), Word32, byteSwap64)
-
-#if WORD_SIZE_IN_BITS < 64
-import GHC.IntWord64
-#endif
+import GHC.Word (Word32, Word64, byteSwap64)
 
 import Numeric (showHex)
 
 import Data.Primitive.Types (Prim (..), defaultSetByteArray#, defaultSetOffAddr#)
 
+import Data.Hashable (Hashable, hashWithSalt)
+import Data.Binary (Binary (get, put))
+
 data Word128 = Word128
   { word128Hi64 :: !Word64
   , word128Lo64 :: !Word64
   }
   deriving (Eq, Data, Generic, Ix, Typeable)
 
+instance Hashable Word128 where
+  hashWithSalt s (Word128 a1 a2) = s `hashWithSalt` a1 `hashWithSalt` a2
+
+-- | @since 0.1.5.0
+instance Binary Word128 where
+  put (Word128 a1 a2) = put a1 >> put a2
+  get = Word128 <$> get <*> get
+
 byteSwapWord128 :: Word128 -> Word128
 byteSwapWord128 (Word128 a1 a0) = Word128 (byteSwap64 a0) (byteSwap64 a1)
 
@@ -133,6 +140,7 @@ instance FiniteBits Word128 where
 instance Real Word128 where
   toRational x = toInteger128 x % 1
 
+-- For unsigned values, quotRem is the same as divMod.
 instance Integral Word128 where
   quot n d = fst (quotRem128 n d)
   rem n d = snd (quotRem128 n d)
@@ -182,17 +190,41 @@ instance Prim Word128 where
 {-# RULES
 "fromIntegral :: Word128 -> Word128" fromIntegral = id :: Word128 -> Word128
 
-"fromIntegral :: Int -> Word128"     fromIntegral = Word128 0 . (fromIntegral :: Int -> Word64)
-"fromIntegral :: Word -> Word128"    fromIntegral = Word128 0 . (fromIntegral :: Word -> Word64)
-"fromIntegral :: Word32 -> Word128"  fromIntegral = Word128 0 . (fromIntegral :: Word32 -> Word64)
+"fromIntegral :: Int -> Word128"     fromIntegral = fromInt
+"fromIntegral :: Word -> Word128"    fromIntegral = fromWord
+"fromIntegral :: Word32 -> Word128"  fromIntegral = fromWord32
 "fromIntegral :: Word64 -> Word128"  fromIntegral = Word128 0
 
-"fromIntegral :: Word128 -> Int"     fromIntegral = \(Word128 _ w) -> fromIntegral w :: Int
-"fromIntegral :: Word128 -> Word"    fromIntegral = \(Word128 _ w) -> fromIntegral w :: Word
-"fromIntegral :: Word128 -> Word32"  fromIntegral = \(Word128 _ w) -> fromIntegral w :: Word32
+"fromIntegral :: Word128 -> Int"     fromIntegral = toInt
+"fromIntegral :: Word128 -> Word"    fromIntegral = toWord
+"fromIntegral :: Word128 -> Word32"  fromIntegral = toWord32
 "fromIntegral :: Word128 -> Word64"  fromIntegral = \(Word128 _ w) -> w
   #-}
 
+{-# INLINE fromInt #-}
+fromInt :: Int -> Word128
+fromInt = Word128 0 . fromIntegral
+
+{-# INLINE fromWord #-}
+fromWord :: Word -> Word128
+fromWord = Word128 0 . fromIntegral
+
+{-# INLINE fromWord32 #-}
+fromWord32 :: Word32 -> Word128
+fromWord32 = Word128 0 . fromIntegral
+
+{-# INLINE toInt #-}
+toInt :: Word128 -> Int
+toInt (Word128 _ w) = fromIntegral w
+
+{-# INLINE toWord #-}
+toWord :: Word128 -> Word
+toWord (Word128 _ w) = fromIntegral w
+
+{-# INLINE toWord32 #-}
+toWord32 :: Word128 -> Word32
+toWord32 (Word128 _ w) = fromIntegral w
+
 -- -----------------------------------------------------------------------------
 -- Functions for `Ord` instance.
 
@@ -232,42 +264,42 @@ fromEnum128 (Word128 _ a0) = fromEnum a0
 
 {-# INLINABLE plus128 #-}
 plus128 :: Word128 -> Word128 -> Word128
-plus128 (Word128 (W64# a1) (W64# a0)) (Word128 (W64# b1) (W64# b0)) =
-  Word128 (W64# s1) (W64# s0)
+plus128 (Word128 a1 a0) (Word128 b1 b0) =
+    Word128 s1 s0
   where
-    !(# c1, s0 #) = plusWord2# a0 b0
-    s1a = plusWord# a1 b1
-    s1 = plusWord# c1 s1a
+    !(c1, s0) = plusCarrySum a0 b0
+    !s1 = a1 + b1 + c1
 
 {-# INLINABLE minus128 #-}
 minus128 :: Word128 -> Word128 -> Word128
-minus128 (Word128 (W64# a1) (W64# a0)) (Word128 (W64# b1) (W64# b0)) =
-  Word128 (W64# d1) (W64# d0)
+minus128 (Word128 a1 a0) (Word128 b1 b0) =
+    Word128 d1 d0
   where
-    !(# d0, c1 #) = subWordC# a0 b0
-    a1c = minusWord# a1 (int2Word# c1)
-    d1 = minusWord# a1c b1
+    !(c1, d0) = subCarryDiff a0 b0
+    !d1 = a1 - c1 - b1
 
 times128 :: Word128 -> Word128 -> Word128
-times128 (Word128 (W64# a1) (W64# a0)) (Word128 (W64# b1) (W64# b0)) =
-  Word128 (W64# p1) (W64# p0)
+times128 (Word128 a1 a0) (Word128 b1 b0) =
+    Word128 p1 p0
   where
-    !(# c1, p0 #) = timesWord2# a0 b0
-    p1a = timesWord# a1 b0
-    p1b = timesWord# a0 b1
-    p1c = plusWord# p1a p1b
-    p1 = plusWord# p1c c1
+    !(c1, p0) = timesCarryProd a0 b0
+    !p1a = a1 * b0
+    !p1b = a0 * b1
+    !p1c = p1a + p1b
+    !p1 = p1c + c1
 
 {-# INLINABLE negate128 #-}
 negate128 :: Word128 -> Word128
-negate128 (Word128 (W64# a1) (W64# a0)) =
-  case plusWord2# (not# a0) 1## of
-    (# c, s #) -> Word128 (W64# (plusWord# (not# a1) c)) (W64# s)
+negate128 (Word128 a1 a0) =
+  case plusCarrySum (complement a0) 1 of
+    (c, s) -> Word128 (complement a1 + c) s
 
 {-# INLINABLE signum128 #-}
 signum128 :: Word128 -> Word128
-signum128 (Word128 (W64# 0##) (W64# 0##)) = zeroWord128
-signum128 _ = oneWord128
+signum128 (Word128 a b) =
+  if a == 0 && b == 0
+    then zeroWord128
+    else oneWord128
 
 fromInteger128 :: Integer -> Word128
 fromInteger128 i =
@@ -278,18 +310,15 @@ fromInteger128 i =
 
 {-# INLINABLE and128 #-}
 and128 :: Word128 -> Word128 -> Word128
-and128 (Word128 (W64# a1) (W64# a0)) (Word128 (W64# b1) (W64# b0)) =
-  Word128 (W64# (and# a1 b1)) (W64# (and# a0 b0))
+and128 (Word128 a1 a0) (Word128 b1 b0) = Word128 (a1 .&. b1) (a0 .&. b0)
 
 {-# INLINABLE or128 #-}
 or128 :: Word128 -> Word128 -> Word128
-or128 (Word128 (W64# a1) (W64# a0)) (Word128 (W64# b1) (W64# b0)) =
-  Word128 (W64# (or# a1 b1)) (W64# (or# a0 b0))
+or128 (Word128 a1 a0) (Word128 b1 b0) = Word128 (a1 .|. b1) (a0 .|. b0)
 
 {-# INLINABLE xor128 #-}
 xor128 :: Word128 -> Word128 -> Word128
-xor128 (Word128 (W64# a1) (W64# a0)) (Word128 (W64# b1) (W64# b0)) =
-  Word128 (W64# (xor# a1 b1)) (W64# (xor# a0 b0))
+xor128 (Word128 a1 a0) (Word128 b1 b0) = Word128 (xor a1 b1) (xor a0 b0)
 
 {-# INLINABLE complement128 #-}
 complement128 :: Word128 -> Word128
@@ -325,8 +354,8 @@ shiftR128 w@(Word128 a1 a0) s
 
 rotateL128 :: Word128 -> Int -> Word128
 rotateL128 w@(Word128 a1 a0) r
-  | r < 0 = zeroWord128
   | r == 0 = w
+  | r < 0 = zeroWord128
   | r >= 128 = rotateL128 w (r `mod` 128)
   | r == 64 = Word128 a0 a1
   | r > 64 = rotateL128 (Word128 a0 a1) (r `mod` 64)
@@ -338,8 +367,8 @@ rotateL128 w@(Word128 a1 a0) r
 
 rotateR128 :: Word128 -> Int -> Word128
 rotateR128 w@(Word128 a1 a0) r
-  | r < 0 = rotateR128 w (128 - (abs r `mod` 128))
   | r == 0 = w
+  | r < 0 = rotateR128 w (128 - (abs r `mod` 128))
   | r >= 128 = rotateR128 w (r `mod` 128)
   | r == 64 = Word128 a0 a1
   | r > 64 = rotateR128 (Word128 a0 a1) (r `mod` 64)
@@ -411,31 +440,24 @@ quotRemFour num@(Word128 n1 _) den@(Word128 d1 _)
 
 {-# INLINE halfTimes128 #-}
 halfTimes128 :: Word128 -> Word64 -> Word128
-halfTimes128 (Word128 (W64# a1) (W64# a0)) (W64# b0) =
-  Word128 (W64# p1) (W64# p0)
+halfTimes128 (Word128 a1 a0) b0 =
+  Word128 p1 p0
   where
-    !(# c1, p0 #) = timesWord2# a0 b0
-    p1a = timesWord# a1 b0
-    p1 = plusWord# p1a c1
+    !(c1, p0) = timesCarryProd a0 b0
+    p1a = a1 * b0
+    p1 = p1a + c1
 
 {-# INLINE quotRemThree #-}
 quotRemThree :: Word128 -> Word64 -> (Word128, Word128)
 quotRemThree num@(Word128 n1 n0) den
   | den == 0 = divZeroError
   | den == 1 = (num, zeroWord128)
-  | n1 < den = case quotRemWord64 n1 n0 den of
+  | n1 < den = case quotRem2Word64 n1 n0 den of
                 (q, r) -> (Word128 0 q, Word128 0 r)
   | otherwise =
       case quotRem n1 den of
-        (q1, r1) -> case quotRemWord64 r1 n0 den of
-                      (q0, r0) -> (Word128 q1 q0, Word128 0 r0)
-
-{-# INLINE quotRemWord64 #-}
-quotRemWord64 :: Word64 -> Word64 -> Word64 -> (Word64, Word64)
-quotRemWord64 (W64# n1) (W64# n0) (W64# d) =
-  case quotRemWord2# n1 n0 d of
-    (# q, r #) -> (W64# q, W64# r)
-
+        (q1, r1) -> case quotRem2Word64 r1 n0 den of
+             (q0, r0) -> (Word128 q1 q0, Word128 0 r0)
 
 {-# INLINE quotRemTwo #-}
 quotRemTwo :: Word64 -> Word64 -> (Word128, Word128)
