@@ -1,6 +1,5 @@
 module Test.Data.WideWord.Gen where
 
-import           Data.Bits (shiftL)
 import           Data.WideWord
 import           Data.Word (Word32, Word64)
 
@@ -19,7 +18,7 @@ genWord32 =
 
 genWord64 :: Gen Word64
 genWord64 =
-  Gen.word64 Range.constantBounded
+  fromIntegral <$> Gen.integral (Range.linear 0 maxBoundWord64)
 
 -- | Generate 'Word64' in one of five categories;
 --    * the full range
@@ -29,20 +28,25 @@ genWord64 =
 --    * values near maxBound :: Word32
 genBiasedWord64 :: Gen Word64
 genBiasedWord64 =
+  fromIntegral <$>
   Gen.choice
-    [ Gen.word64 (Range.linear 0 maxBound)
-    , Gen.word64 (Range.linear 0 100)
-    , (-) maxBound <$> Gen.word64 (Range.linear 0 100)
-    , Gen.word64 (Range.linear (halfMax - 100) (halfMax + 100))
-    , Gen.word64 (Range.linear (bits32 - 100) (bits32 + 100))
+    [ Gen.integral (Range.linear 0 maxBoundWord64)
+    , Gen.integral (Range.linear 0 100)
+    , (-) maxBoundWord64 <$> Gen.integral (Range.linear 0 100)
+    , Gen.integral (Range.linear (halfMax - 100) (halfMax + 100))
+    , Gen.integral (Range.linear (bits32 - 100) (bits32 + 100))
     ]
   where
-    bits32 :: Word64
-    bits32 = 1 `shiftL` 32
+    bits32 :: Integer
+    bits32 = fromIntegral (maxBound :: Word32)
 
-    halfMax :: Word64
-    halfMax = maxBound `div` 2
+    halfMax :: Integer
+    halfMax = fromIntegral (maxBound `div` 2 :: Word64)
 
 genWord128 :: Gen Word128
 genWord128 =
   Word128 <$> genBiasedWord64 <*> genBiasedWord64
+
+maxBoundWord64 :: Integer
+maxBoundWord64 = fromIntegral (maxBound :: Word64)
+
